<?php

namespace Drupal\oidc\EventSubscriber;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\externalauth\Event\ExternalAuthRegisterEvent;
use Drupal\oidc\OpenidConnectSessionInterface;
use Drupal\externalauth\Event\ExternalAuthEvents;
use Drupal\externalauth\Event\ExternalAuthLoginEvent;
use Drupal\oidc\Plugin\OpenidConnectRealm\GenericOpenidConnectRealm;
use Drupal\user\UserInterface;
use InvalidArgumentException;
use Symfony\Component\EventDispatcher\Event;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Event subscriber for External Authentication events.
 */
class ExternalAuthEventSubscriber implements EventSubscriberInterface {

  /**
   * The OpenID Connect session service.
   *
   * @var \Drupal\oidc\OpenidConnectSessionInterface
   */
  protected $session;

  /**
   * The user storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $userStorage;

  /**
   * Class constructor.
   *
   * @param \Drupal\oidc\OpenidConnectSessionInterface $session
   *   The OpenID Connect session service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(OpenidConnectSessionInterface $session, EntityTypeManagerInterface $entity_type_manager) {
    $this->session = $session;
    $this->userStorage = $entity_type_manager->getStorage('user');
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events[ExternalAuthEvents::REGISTER][] = ['onRegister', 1000];
    $events[ExternalAuthEvents::LOGIN][] = ['onLogin', 1000];

    return $events;
  }

  /**
   * Assigns a default role when a user is registered.
   *
   * @param \Drupal\externalauth\Event\ExternalAuthRegisterEvent $event
   *   The register event.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function onRegister(ExternalAuthRegisterEvent $event) {
    if (!$this->isActiveOpenidConnectRealm($event)) {
      return;
    }

    // This is only possible for generic realms.
    $plugin = $this->session->getRealmPlugin();
    if (!$plugin instanceof GenericOpenidConnectRealm) {
      return;
    }

    // Assign the default role.
    if (($rid = $plugin->getDefaultRoleId()) !== NULL) {
      $account = $event->getAccount();
      $account->addRole($rid);
      $account->save();
    }
  }

  /**
   * Updates the username, email, given name and family name when a user logs in.
   *
   * @param \Drupal\externalauth\Event\ExternalAuthLoginEvent $event
   *   The login event.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function onLogin(ExternalAuthLoginEvent $event) {
    if (!$this->isActiveOpenidConnectRealm($event)) {
      return;
    }

    $tokens = $this->session->getJsonWebTokens();
    $account = $event->getAccount();
    $save = FALSE;

    // Update the username.
    $username = $tokens->getUsername();

    if ($username !== NULL) {
      $username = mb_substr($username, 0, UserInterface::USERNAME_MAX_LENGTH);
      $account_name = $account->getAccountName();

      if ($username !== $account_name) {
        $regex = mb_substr($username, 0, UserInterface::USERNAME_MAX_LENGTH - 2);
        $regex = '#^' .  preg_quote($regex, '#') . '_\d$#';

        if (!preg_match($regex, $account_name)) {
          $username = $this->generateUsername($username);

          if ($username !== NULL) {
            $account->setUsername($username);
            $save = TRUE;
          }
        }
      }
    }

    // Update the email address.
    $email = $tokens->getEmail();

    if ($email !== NULL && $email !== $account->getEmail()) {
      $account->setEmail($email);
      $save = TRUE;
    }

    // Update the given name.
    $given_name = $tokens->getGivenName();

    if ($given_name !== NULL && $given_name !== $account->get('given_name')->value) {
      $account->set('given_name', $given_name);
      $save = TRUE;
    }

    // Update the family name.
    $family_name = $tokens->getFamilyName();

    if ($family_name !== NULL && $family_name !== $account->get('family_name')->value) {
      $account->set('family_name', $family_name);
      $save = TRUE;
    }

    // Save the changes.
    if ($save) {
      $account->save();
    }
  }

  /**
   * Checks if the event provider is the active OpenID Connect realm.
   *
   * @param \Drupal\externalauth\Event\ExternalAuthRegisterEvent|\Drupal\externalauth\Event\ExternalAuthLoginEvent $event
   *   The external authentication event.
   *
   * @return bool
   *   TRUE if the provider matches the active OpenID Connect realm.
   */
  protected function isActiveOpenidConnectRealm(Event $event) {
    if (!$event instanceof ExternalAuthRegisterEvent && !$event instanceof ExternalAuthLoginEvent) {
      throw new InvalidArgumentException('Specified event is no external authentication event.');
    }

    if (!$plugin_id = $this->session->getRealmPluginId()) {
      return FALSE;
    }

    return $event->getProvider() === 'oidc:' . $plugin_id;
  }

  /**
   * Try to generate a unique username.
   *
   * @param string $base
   *   The string to use as base.
   *
   * @return string|null
   *   The username or null if no unique username could be generated.
   */
  protected function generateUsername($base) {
    // Check if $base exists, otherwise just use that.
    $count = $this->userStorage->getQuery()
      ->count()
      ->condition('name', $base)
      ->execute();

    if (!$count) {
      return $base;
    }

    // Try suffix "_1" until "_9".
    $base = mb_substr($base, 0, UserInterface::USERNAME_MAX_LENGTH - 2);

    for ($index = 1; $index <= 9; $index++) {
      $name = $base . '_' . $index;
      $count = $this->userStorage->getQuery()
        ->count()
        ->condition('name', $name)
        ->execute();

      if (!$count) {
        return $name;
      }
    }
  }

}
