<?php

namespace Drupal\oidc;

/**
 * Holds the JSON web tokens.
 */
class JsonWebTokens {

  /**
   * The token type.
   *
   * @var string
   */
  protected $type;

  /**
   * The ID token.
   *
   * @var string
   */
  protected $idToken;

  /**
   * The access token.
   *
   * @var string
   */
  protected $accessToken;

  /**
   * Unix timestamp when the token expires.
   *
   * @var int
   */
  protected $expires;

  /**
   * Associative array of claims.
   *
   * @var array
   */
  protected $claims = [];

  /**
   * Name of the ID claim.
   *
   * @var string
   */
  protected $idClaim = 'sub';

  /**
   * Name of the username claim.
   *
   * @var string
   */
  protected $usernameClaim = 'preferred_username';

  /**
   * Name of the email claim.
   *
   * @var string
   */
  protected $emailClaim = 'email';

  /**
   * Name of the given name claim.
   *
   * @var string
   */
  protected $givenNameClaim = 'given_name';

  /**
   * Name of the family name claim.
   *
   * @var string
   */
  protected $familyNameClaim = 'family_name';

  /**
   * Class constructor.
   *
   * @param string $type
   *   The token type.
   * @param string $id_token
   *   The ID token.
   * @param string $acces_token
   *   The access token.
   * @param int $expires
   *   Unix timestamp when the token expires.
   */
  public function __construct($type, $id_token, $acces_token, $expires) {
    $this->type = $type;
    $this->idToken = $id_token;
    $this->accessToken = $acces_token;
    $this->expires = $expires;
  }

  /**
   * Create an instance based on an array.
   *
   * @param array $data
   *   The array containing all object properties.
   *
   * @return static
   *   The class instance.
   */
  public static function fromArray(array $data) {
    $instance = new static(
      $data['type'],
      $data['id_token'],
      $data['access_token'],
      $data['expires']
    );

    $instance->claims = $data['claims'];
    $instance->idClaim = $data['id_claim'];
    $instance->usernameClaim = $data['username_claim'];
    $instance->emailClaim = $data['email_claim'];
    $instance->givenNameClaim = $data['given_name_claim'];
    $instance->familyNameClaim = $data['family_name_claim'];

    return $instance;
  }

  /**
   * Get the ID token.
   *
   * @return string
   *   The ID token.
   */
  public function getIdToken() {
    return $this->idToken;
  }

  /**
   * Get the access token.
   *
   * @return string
   *   The access token.
   */
  public function getAccessToken() {
    return $this->accessToken;
  }

  /**
   * Get the expires timestamp.
   *
   * @return int
   *   The expires timestamp.
   */
  public function getExpires() {
    return $this->expires;
  }

  /**
   * Get an associative array of all claims.
   *
   * @return array
   *   The claims.
   */
  public function getClaims() {
    return $this->claims;
  }

  /**
   * Get a single claim.
   *
   * @param string $claim
   *   The claim.
   *
   * @return string|int|array|null
   *   The claim value or NULL if missing.
   */
  public function getClaim($claim) {
    if (array_key_exists($claim, $this->claims)) {
      return $this->claims[$claim];
    }

    return NULL;
  }

  /**
   * Set a claim.
   *
   * @param string $claim
   *   The claim.
   * @param string|int|array $value
   *   The claim value.
   *
   * @return $this
   *   The class instance.
   */
  public function setClaim($claim, $value) {
    $this->claims[$claim] = $value;

    return $this;
  }

  /**
   * Get the ID claim.
   *
   * @return int|string|null
   *   The ID or NULL if missing.
   */
  public function getId() {
    return $this->getClaim($this->idClaim);
  }

  /**
   * Get the username claim.
   *
   * @return int|string|null
   *   The username or NULL if missing.
   */
  public function getUsername() {
    return $this->getClaim($this->usernameClaim);
  }

  /**
   * Get the email claim.
   *
   * @return int|string|null
   *   The email address or NULL if missing.
   */
  public function getEmail() {
    return $this->getClaim($this->emailClaim);
  }

  /**
   * Get the given name claim.
   *
   * @return int|string|null
   *   The given name or NULL if missing.
   */
  public function getGivenName() {
    return $this->getClaim($this->givenNameClaim);
  }

  /**
   * Get the family name claim.
   *
   * @return int|string|null
   *   The family name or NULL if missing.
   */
  public function getFamilyName() {
    return $this->getClaim($this->familyNameClaim);
  }

  /**
   * Set the name of the ID claim.
   *
   * @param string $claim
   *   The claim name.
   *
   * @return $this
   *   The class instance.
   */
  public function setIdClaim($claim) {
    $this->idClaim = $claim;

    return $this;
  }

  /**
   * Set the name of the username claim.
   *
   * @param string $claim
   *   The claim name.
   *
   * @return $this
   *   The class instance.
   */
  public function setUsernameClaim($claim) {
    $this->usernameClaim = $claim;

    return $this;
  }

  /**
   * Set the name of the email address claim.
   *
   * @param string $claim
   *   The claim name.
   *
   * @return $this
   *   The class instance.
   */
  public function setEmailClaim($claim) {
    $this->emailClaim = $claim;

    return $this;
  }

  /**
   * Set the name of the given name claim.
   *
   * @param string $claim
   *   The claim name.
   *
   * @return $this
   *   The class instance.
   */
  public function setGivenNameClaim($claim) {
    $this->givenNameClaim = $claim;

    return $this;
  }

  /**
   * Set the name of the family name claim.
   *
   * @param string $claim
   *   The claim name.
   *
   * @return $this
   *   The class instance.
   */
  public function setFamilyNameClaim($claim) {
    $this->familyNameClaim = $claim;

    return $this;
  }

  /**
   * Get the array representation of this object.
   *
   * @return array
   *   The object properties as array.
   */
  public function toArray() {
    return [
      'type' => $this->type,
      'id_token' => $this->idToken,
      'access_token' => $this->accessToken,
      'expires' => $this->expires,
      'claims' => $this->claims,
      'id_claim' => $this->idClaim,
      'username_claim' => $this->usernameClaim,
      'email_claim' => $this->emailClaim,
      'given_name_claim' => $this->givenNameClaim,
      'family_name_claim' => $this->familyNameClaim,
    ];
  }

}
